
import Test.QuickCheck
import Data.Char( digitToInt )

--------------------------------------------------------------------------------

data List a
  = Nil
  | Cons a (List a)
 deriving ( Eq, Show )

instance Arbitrary a => Arbitrary (List a) where
  arbitrary =
    frequency
    [ (1, do return Nil)
    , (5, do x <- arbitrary
             xs <- arbitrary
             return (Cons x xs))
    ]

--------------------------------------------------------------------------------

data Tree a
  = Node a (Tree a) (Tree a)
  | Empty
 deriving ( Eq, Show )

instance Arbitrary a => Arbitrary (Tree a) where
  arbitrary = sized arbTree
   where
    arbTree s =
      frequency
      [ (1, do return Empty)
      , (s, do x <- arbitrary
               p <- arbTree (s `div` 2)
               q <- arbTree (s `div` 2)
               return (Node x p q))
      ]

--------------------------------------------------------------------------------

data Expr
  = Add Expr Expr
  | Mul Expr Expr
  | Num Int
  | Var Char
 deriving ( Eq, Show )

instance Arbitrary Expr where
  arbitrary = sized arbExpr
   where
    arbExpr s =
      frequency
      [ (1, do v <- choose ('a','z')
               return (Var v))
      , (1, do n <- arbitrary
               return (Num n))
      , (s, do a <- arbExpr (s `div` 2)
               b <- arbExpr (s `div` 2)
               return (Add a b))
      , (s, do a <- arbExpr (s `div` 2)
               b <- arbExpr (s `div` 2)
               return (Mul a b))
      ]

--------------------------------------------------------------------------------

type PersonNumber = [Int]

arbPersonNumber :: Gen PersonNumber
arbPersonNumber =
  do y <- choose (0,99)
     m <- choose (1,12)
     d <- choose (1,daysIn m)
     a <- choose (0,9999)
     return (map digitToInt (pad 2 y ++ pad 2 m ++ pad 2 d ++ pad 4 a))
 where
  daysIn 2                       = 28 -- ignoring leap year...
  daysIn m | m `elem` [4,6,9,11] = 30
           | otherwise           = 31
  
  pad k x = replicate (k - length s) '0' ++ s
   where
    s = show (x :: Int)









