{-# LANGUAGE ScopedTypeVariables #-}
import Test.QuickCheck hiding ( OrderedList )
import Data.List ( (\\), sort )

--------------------------------------------------------------

prop_Ordered (xs :: [Int]) =
  ordered (msort xs)

ordered :: Ord a => [a] -> Bool
ordered []       = True
ordered [x]      = True
ordered (x:y:xs) = x <= y && ordered (y:xs)

--------------------------------------------------------------

-- this property becomes obsolete after we add the next property

prop_Lengths (xs :: [Int]) =
  length (msort xs) == length xs

--------------------------------------------------------------

prop_Permutation (xs :: [Int]) =
  msort xs `isPermutation` xs

[]     `isPermutation` ys = null ys
(x:xs) `isPermutation` ys = x `elem` ys && xs `isPermutation` (ys \\ [x])

--------------------------------------------------------------

msort :: Ord a => [a] -> [a]
msort xs = merging (map unit xs)
 where
  unit x = [x]

merging :: Ord a => [[a]] -> [a]
merging []   = []
merging [xs] = xs
merging xss  = merging (pairwise xss)

pairwise :: Ord a => [[a]] -> [[a]]
pairwise (xs:ys:xss) = merge xs ys : pairwise xss
pairwise xss         = xss

merge :: Ord a => [a] -> [a] -> [a]
merge []     ys     = ys
merge xs     []     = xs
merge (x:xs) (y:ys)
  | x <= y          = x : merge xs (y:ys)
  | otherwise       = y : merge (x:xs) ys

{-
-- buggy version:
merge (x:xs) (y:ys)
  | x <= y          = x : merge xs ys
  | otherwise       = y : merge xs ys
-}

--------------------------------------------------------------

prop_MergePermute xs (ys :: [Int]) =
  merge xs ys `isPermutation` (xs ++ ys)
  
prop_MergeOrdered (MyList xs) (MyList (ys :: [Int])) =
  collect (length xs) $
    ordered (merge xs ys)

{-
-- bad version that is hard to test:
prop_MergeOrdered xs (ys :: [Int]) =
  collect (length xs) $
    ordered xs && ordered ys ==>
      ordered (merge xs ys)
-}

-- generator for ordered lists

data OrderedList a = MyList [a]
 deriving ( Eq, Show )

instance (Ord a, Arbitrary a) => Arbitrary (OrderedList a) where
  arbitrary =
    do xs <- arbitrary
       return (MyList (sort xs))
  
--------------------------------------------------------------

